import base64
import hashlib
import os

from zeep import ns
from zeep.wsse import utils


class UsernameToken:
    """UsernameToken Profile 1.1

    https://docs.oasis-open.org/wss/v1.1/wss-v1.1-spec-os-UsernameTokenProfile.pdf

    Example response using PasswordText::

        <wsse:Security>
          <wsse:UsernameToken>
            <wsse:Username>scott</wsse:Username>
            <wsse:Password Type="wsse:PasswordText">password</wsse:Password>
          </wsse:UsernameToken>
        </wsse:Security>

    Example using PasswordDigest::

        <wsse:Security>
          <wsse:UsernameToken>
            <wsse:Username>NNK</wsse:Username>
            <wsse:Password Type="wsse:PasswordDigest">
                weYI3nXd8LjMNVksCKFV8t3rgHh3Rw==
            </wsse:Password>
            <wsse:Nonce>WScqanjCEAC4mQoBE07sAQ==</wsse:Nonce>
            <wsu:Created>2003-07-16T01:24:32Z</wsu:Created>
          </wsse:UsernameToken>
        </wsse:Security>

    """

    username_token_profile_ns = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0"  # noqa
    soap_message_secutity_ns = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0"  # noqa

    def __init__(
        self,
        username,
        password=None,
        password_digest=None,
        use_digest=False,
        nonce=None,
        created=None,
        timestamp_token=None,
        zulu_timestamp=None,
        hash_password=None,
    ):
        """
        Some SOAP services want zulu timestamps with Z in timestamps and
        in password digests they may want password to be hashed before
        adding it to nonce and created.
        """
        self.username = username
        self.password = password
        self.password_digest = password_digest
        self.nonce = nonce
        self.created = created
        self.use_digest = use_digest
        self.timestamp_token = timestamp_token
        self.zulu_timestamp = zulu_timestamp
        self.hash_password = hash_password

    def apply(self, envelope, headers):
        security = utils.get_security_header(envelope)

        # The token placeholder might already exists since it is specified in
        # the WSDL.
        token = security.find("{%s}UsernameToken" % ns.WSSE)
        if token is None:
            token = utils.WSSE.UsernameToken()
            security.append(token)

        if self.timestamp_token is not None:
            security.append(self.timestamp_token)

        # Create the sub elements of the UsernameToken element
        elements = [utils.WSSE.Username(self.username)]
        if self.password is not None or self.password_digest is not None:
            if self.use_digest:
                elements.extend(self._create_password_digest())
            else:
                elements.extend(self._create_password_text())

        token.extend(elements)
        return envelope, headers

    def verify(self, envelope):
        pass

    def _create_password_text(self):
        return [
            utils.WSSE.Password(
                self.password, Type="%s#PasswordText" % self.username_token_profile_ns
            )
        ]

    def _create_password_digest(self):
        if self.nonce:
            nonce = self.nonce.encode("utf-8")
        else:
            nonce = os.urandom(16)
        timestamp = utils.get_timestamp(self.created, self.zulu_timestamp)

        if isinstance(self.password, str):
            password = self.password.encode("utf-8")
        else:
            password = self.password

        # digest = Base64 ( SHA-1 ( nonce + created + password ) )
        if not self.password_digest and self.hash_password:
            digest = base64.b64encode(
                hashlib.sha1(
                    nonce + timestamp.encode("utf-8") + hashlib.sha1(password).digest()
                ).digest()
            ).decode("ascii")
        elif not self.password_digest:
            digest = base64.b64encode(
                hashlib.sha1(nonce + timestamp.encode("utf-8") + password).digest()
            ).decode("ascii")
        else:
            digest = self.password_digest

        return [
            utils.WSSE.Password(
                digest, Type="%s#PasswordDigest" % self.username_token_profile_ns
            ),
            utils.WSSE.Nonce(
                base64.b64encode(nonce).decode("utf-8"),
                EncodingType="%s#Base64Binary" % self.soap_message_secutity_ns,
            ),
            utils.WSU.Created(timestamp),
        ]



"""Basic implementation to support SOAP-Attachments

See https://www.w3.org/TR/SOAP-attachments

"""

import base64

from cached_property import cached_property
from requests.structures import CaseInsensitiveDict


class MessagePack:
    def __init__(self, parts):
        self._parts = parts

    def __repr__(self):
        return "<MessagePack(attachments=[%s])>" % (
            ", ".join(repr(a) for a in self.attachments)
        )

    @property
    def root(self):
        return self._root

    def _set_root(self, root):
        self._root = root

    @cached_property
    def attachments(self):
        """Return a list of attachments.

        :rtype: list of Attachment

        """
        return [Attachment(part) for part in self._parts]

    def get_by_content_id(self, content_id):
        """get_by_content_id

        :param content_id: The content-id to return
        :type content_id: str
        :rtype: Attachment

        """
        for attachment in self.attachments:
            if attachment.content_id == content_id:
                return attachment


class Attachment:
    def __init__(self, part):
        encoding = part.encoding or "utf-8"
        self.headers = CaseInsensitiveDict(
            {k.decode(encoding): v.decode(encoding) for k, v in part.headers.items()}
        )
        self.content_type = self.headers.get("Content-Type", None)
        self.content_id = self.headers.get("Content-ID", None)
        self.content_location = self.headers.get("Content-Location", None)
        self._part = part

    def __repr__(self):
        return "<Attachment(%r, %r)>" % (self.content_id, self.content_type)

    @cached_property
    def content(self):
        """Return the content of the attachment

        :rtype: bytes or str

        """
        encoding = self.headers.get("Content-Transfer-Encoding", None)
        content = self._part.content

        if encoding == "base64":
            return base64.b64decode(content)
        elif encoding == "binary":
            return content.strip(b"\r\n")
        else:
            return content



from zeep.settings import Settings


class XmlParserContext:
    """Parser context when parsing XML elements"""

    def __init__(self, settings=None):
        self.schemas = []
        self.settings = settings or Settings()



"""
    zeep.wsdl.messages
    ~~~~~~~~~~~~~~~~~~

    The messages are responsible for serializing and deserializing

    .. inheritance-diagram::
            zeep.wsdl.messages.soap.DocumentMessage
            zeep.wsdl.messages.soap.RpcMessage
            zeep.wsdl.messages.http.UrlEncoded
            zeep.wsdl.messages.http.UrlReplacement
            zeep.wsdl.messages.mime.MimeContent
            zeep.wsdl.messages.mime.MimeXML
            zeep.wsdl.messages.mime.MimeMultipart
       :parts: 1

"""
from .http import *  # noqa
from .mime import *  # noqa
from .soap import *  # noqa



import re

from lxml import etree


def process_multiref(node):
    """Iterate through the tree and replace the referened elements.

    This method replaces the nodes with an href attribute and replaces it
    with the elements it's referencing to (which have an id attribute).abs

    """
    multiref_objects = {elm.attrib["id"]: elm for elm in node.xpath("*[@id]")}
    if not multiref_objects:
        return

    used_nodes = []

    def process(node):
        """Recursive"""
        # TODO (In Soap 1.2 this is 'ref')
        href = node.attrib.get("href")

        if href and href.startswith("#"):
            obj = multiref_objects.get(href[1:])
            if obj is not None:
                used_nodes.append(obj)
                node = _dereference_element(obj, node)

        for child in node:
            process(child)

    process(node)

    # Remove the old dereferenced nodes from the tree
    for node in used_nodes:
        parent = node.getparent()
        if parent is not None:
            parent.remove(node)


def _dereference_element(source, target):
    """Move the referenced node (source) in the main response tree (target)

    :type source: lxml.etree._Element
    :type target: lxml.etree._Element
    :rtype target: lxml.etree._Element

    """
    specific_nsmap = {k: v for k, v in source.nsmap.items() if k not in target.nsmap}

    new = _clone_element(source, target.tag, specific_nsmap)

    # Replace the node with the new dereferenced node
    parent = target.getparent()
    parent.insert(parent.index(target), new)
    parent.remove(target)

    # Update all descendants
    for obj in new.iter():
        _prefix_node(obj)

    return new


def _clone_element(node, tag_name=None, nsmap=None):
    """Clone the given node and return it.

    This is a recursive call since we want to clone the children the same
    way.

    :type source: lxml.etree._Element
    :type tag_name: str
    :type nsmap: dict
    :rtype source: lxml.etree._Element

    """
    tag_name = tag_name or node.tag
    nsmap = node.nsmap if nsmap is None else nsmap
    new = etree.Element(tag_name, nsmap=nsmap)

    for child in node:
        new_child = _clone_element(child)
        new.append(new_child)
    new.text = node.text

    for key, value in _get_attributes(node):
        new.set(key, value)

    return new


def _prefix_node(node):
    """Translate the internal attribute values back to prefixed tokens.

    This reverses the translation done in _get_attributes

    For example::

        {
            'foo:type': '{http://example.com}string'
        }

    will be converted to:

        {
            'foo:type': 'example:string'
        }

    :type node: lxml.etree._Element

    """
    reverse_nsmap = {v: k for k, v in node.nsmap.items()}

    prefix_re = re.compile("^{([^}]+)}(.*)")

    for key, value in node.attrib.items():
        if value.startswith("{"):
            match = prefix_re.match(value)
            if not match:
                continue
            namespace, localname = match.groups()

            if namespace in reverse_nsmap:
                value = "%s:%s" % (reverse_nsmap.get(namespace), localname)
                node.set(key, value)


def _get_attributes(node):
    """Return the node attributes where prefixed values are dereferenced.

    For example the following xml::

        <foobar xmlns:xsi="foo" xmlns:ns0="bar" xsi:type="ns0:string">

    will return the dict::

        {
            'foo:type': '{http://example.com}string'
        }

    :type node: lxml.etree._Element

    """
    nsmap = node.nsmap
    result = {}

    for key, value in node.attrib.items():
        if value.count(":") == 1:
            prefix, localname = value.split(":")

            if prefix in nsmap:
                namespace = nsmap[prefix]
                value = "{%s}%s" % (namespace, localname)
        result[key] = value
    return list(result.items())
